/*
 * Copyright (c) 2021, Xilinx Inc. and Contributors. All rights reserved.
 *
 * SPDX-License-Identifier: MIT
 */

/*
 * @file	assert.h
 * @brief	assert definition.
 */

#ifndef _ACAPD_ASSERT_H
#define _ACAPD_ASSERT_H
#include <assert.h>

#define _unused(x) ((void)(x))

#ifdef __cplusplus
extern "C" {
#endif

void acapd_assert(unsigned int cond);

#ifdef __cplusplus
}
#endif

#endif /*  _ACAPD_COMMON_H */
