/*
 * Copyright (c) 2021, Xilinx Inc. and Contributors. All rights reserved.
 *
 * SPDX-License-Identifier: MIT
 */

/*
 * @file	linux/dma.h
 * @brief	Linux specific DMA definition.
 */


#ifndef _ACAPD_LINUX_DMA_H
#define _ACAPD_LINUX_DMA_H

extern acapd_dma_ops_t axidma_vfio_dma_ops;
extern acapd_dma_ops_t mcdma_vfio_dma_ops;

#endif /* _ACAPD_LINUX_DMA_H */
